import shutil
import random
import numpy as np
import matplotlib.pyplot as plot
from scipy.io import wavfile
import soundfile as sf
import subprocess
import time
from skimage.metrics import structural_similarity
import cv2
import os
from PIL import Image
import image_slicer
from imageai.Detection.Custom import CustomObjectDetection

initfolder = 'init'
nodeoffset=27 #audio output node nr in init pd-file
puredataexepath='C:\\Program Files\\Pd\\bin\\pd.exe'

detector = CustomObjectDetection()
detector.setModelTypeAsYOLOv3()
detector.setModelPath("detection_model-ex-012--loss-0026.467.h5") 
detector.setJsonPath("detection_config.json")
detector.loadModel()

def OpenPdFile(filepath):
    process = subprocess.Popen([puredataexepath,filepath])
    time.sleep(21)
    process.kill()

def limit(n, minn, maxn):
    return max(min(maxn, n), minn)

def SubFolderCount(path):
    count = 0
    for x in os.listdir(path):
        if os.path.isdir(os.path.join(path,x)):
            count = count + 1
    return count

def Nmaxelements(list1, N):
    final_list = []
    for i in range(0, N): 
        max1 = 0
        for j in range(len(list1)):     
            if list1[j] > max1:
                max1 = list1[j];
        list1.remove(max1);
        final_list.append(max1)
    return final_list

def DetectChirps(generation):
    amount = SubFolderCount(str(generation))
    imagename= 'rec.wav-r_01_'
    imageextension='.png'
    FinalScores=[]
    for x in range(amount):
        scores = [[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0]]
        for y in range(12):
            doublednumber = "%02d" % (y+1,)
            ImagePath = str(generation)+"/"+str(x)+"/"+imagename+doublednumber+imageextension
            ImageDumpPath = str(generation)+"/"+str(x)+"/"+doublednumber+".jpg"
            print(ImageDumpPath)
            detections = detector.detectObjectsFromImage(input_image=ImagePath, output_image_path=ImageDumpPath, minimum_percentage_probability = 35)
            for detection in detections:
                scores[int(detection["name"])-1].append(detection["percentage_probability"])
        #print(scores)
        SimpleScores = []
        for z in range(len(scores)):
            SimpleScores.append(Nmaxelements(scores[z], 5))
        #print(SimpleScores)
        AlmostFinalScores = []
        for zz in range(len(SimpleScores)):
            AlmostFinalScores.append(sum(SimpleScores[zz]) / len(SimpleScores[zz]))
        print(AlmostFinalScores)
        FinalScores.append(sum(AlmostFinalScores))
    print(FinalScores)
    with open(str(generation)+"/fitness.txt", "a") as myfile:
        myfile.write(str(FinalScores)) 
            
        


    
def SliceImages(generation):
    amount = SubFolderCount(str(generation))
    imagename= 'rec.wav'
    imageextension='.png'
    for x in range(amount):
        image = Image.open(str(generation)+"/"+str(x)+"/"+imagename+imageextension)
        new_image = image.resize((2784, 204))
        new_image.save(str(generation)+"/"+str(x)+"/"+imagename+"-r"+imageextension)
        image_slicer.slice(str(generation)+"/"+str(x)+"/"+imagename+"-r"+imageextension, row=1, col=12)

#generate blank slate pheno
def NewIt(generation,iteration):
    source_dir = r""+initfolder+""
    destination_dir = r""+str(generation)+"/"+str(iteration)+""
    shutil.copytree(source_dir, destination_dir)
    
def Similarity(original, generated):
  img1 = cv2.imread(original,0)
  img2 = cv2.imread(generated,0)
  orb = cv2.ORB_create()
  kp_a, desc_a = orb.detectAndCompute(img1, None)
  kp_b, desc_b = orb.detectAndCompute(img2, None)
  bf = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)
  matches = bf.match(desc_a, desc_b)
  similar_regions = [i for i in matches if i.distance < 50]  
  if len(matches) == 0:
    return 0
  sim, diff = structural_similarity(img1, img2, full=True)
  #returns [ssim,orb]
  return [sim, len(similar_regions) / len(matches)]

#generate random gene format [ [[node1freq,node1x,node1y,node1waveform][x,x,x,x][x,x,x,x]] [[connect1from,connect1to][x,x][x,x][x,x]] ]
def RandomGene(nodeamt):
    gene = [[],[]]
    possibleconnects = [0]
    for idx,node in enumerate(range(nodeamt)):
        node = [round(abs(np.random.normal(0, 0.1))*abs(np.random.normal(0, 10))*abs(np.random.normal(0, 100)),3),random.randrange(0,127),random.randrange(0,127),random.randrange(0,3)]
        gene[0].append(node)
        connect = [idx+1,random.choice(possibleconnects)]
        gene[1].append(connect)
        possibleconnects.append(idx)
    return gene

#generate phenotype: makes gene into entity (puredatafile)
def GeneratePheno(generation,iteration,gene):
    #generate blank slate
    NewIt(generation,iteration)
    nodelevels = [0]
    for idx,connection in enumerate(gene[1]):
        #write connect to pd file
        leveltoplacenode=nodelevels[connection[1]]+1
        nodelevels.append(leveltoplacenode)
        rowtoplacenode=nodelevels.count(leveltoplacenode)
        with open(""+str(generation)+"/"+str(iteration)+"/main.pd", "a") as myfile:
            myfile.write("\n#X obj "+str(rowtoplacenode*200)+" "+str(600-leveltoplacenode*130)+" basic-lfo "+str(idx+1)+";")
            myfile.write("\n#X connect "+str(connection[0]+nodeoffset)+" 0 "+str(connection[1]+nodeoffset)+" 0;")
    #savestates
    for idx,node in enumerate(gene[0]):
        with open(""+str(generation)+"/"+str(iteration)+"/patch_editor_abs/statesave/"+str(idx+1)+"-basic-lfo-ss.txt", "w+") as myfile:
            myfile.write(str(node[0])+" "+str(node[1])+" "+str(node[2])+" "+str(node[3])+";")


 
def GenerateRandomPopulation(gennumber,amount,complexity):
    for x in range(amount):
        gene=RandomGene(complexity)
        GeneratePheno(gennumber,x,gene)
        with open(str(gennumber)+"/genes.txt", "a") as myfile:
             myfile.write(str(gene))
             myfile.write("\n")


def GenerateSpectrogram(generation,x):
    CurrentItem = str(generation)+"/"+str(x)+"/rec.wav"
    CurrentItemRoot = str(generation)+"/"+str(x)
    f = sf.SoundFile(CurrentItem)
    samplingFrequency, signalData = wavfile.read(CurrentItem)
    widthHeight = (int(f.frames / f.samplerate), 1.5)
    plot.figure(figsize=widthHeight).add_axes([0,0,1,1])
    plot.specgram(signalData,Fs=samplingFrequency, cmap='binary')
    plot.axis('off')
    plot.savefig(CurrentItem+'.png',bbox_inches='tight', dpi=1200)
    plot.savefig(CurrentItemRoot+'.png',bbox_inches='tight', dpi=1200)


def PerformPheno(generation):
    amount = SubFolderCount(str(generation))
    for x in range(amount):
        OpenPdFile(str(generation)+"/"+str(x)+"/main.pd")
        GenerateSpectrogram(generation,x)
        
def GetFitness2(generation):
    SliceImages(generation)
    DetectChirps(generation)

def GetFitness(generation):
    amount = SubFolderCount(str(generation))
    fitness = []
    for x in range(amount):
        sim = Similarity(str(generation)+"/"+str(x)+"/rec.wav.png",'god.png')
        fitness.append(sim)
        
    with open(str(generation)+"/fitness.txt", "a") as myfile:
        myfile.write(str(fitness))
        
def ModifyGeneNodeValues(gene):
    for node,idx in enumerate(gene[0]):
        gene[0][node][0] = round(gene[0][node][0]*random.uniform(0.95, 1.05),3)
        gene[0][node][1] = limit(round(gene[0][node][1]*random.uniform(0.8, 1.2),0),0,126)
        gene[0][node][2] = limit(round(gene[0][node][2]*random.uniform(0.8, 1.2),0),0,126)
        #gene[0][node][3] = gene[0][node][3]*random.uniform(0.9, 1.1)
    return gene

def SexyTimeFrom(parents):
    NewGene = [[],[]]
    for node,idx in enumerate(random.choice(parents)[0]):
        NewGene[0].append(random.choice(parents)[0][node])
        NewGene[1].append(random.choice(parents)[1][node])
    return NewGene

def ChangeRandomConnect(gene):
    changeconnect = random.randrange(1,len(gene[1]))
    gene[1][changeconnect][1]=random.randrange(0,changeconnect)+1
    return gene
    
def Breed(generation):
    amount = SubFolderCount(str(generation))
    #get genes from generation
    AllGenes= []
    with open (str(generation)+"/genes.txt", "r") as myfile:
        data = myfile.read()
        lines = data.split("\n")
        lines.pop()
        for each_line in lines:
            AllGenes.append(eval(each_line))
            
    #get top entities
    AllScores = []
    with open (str(generation)+"/fitness.txt", "r") as myfile:
        data = myfile.read()
        AllScores = eval(data)

    First = AllScores.index(max(AllScores))
    newlist = AllScores
    newlist[newlist.index(max(newlist))]=0
    Second = newlist.index(max(newlist))
    print(str(First) +'&'+str(Second))
    TopEntities = [AllGenes[First],AllGenes[Second]]

    for x in range(amount):
        #sexytime generate new genes from parents
        NewGene=SexyTimeFrom(TopEntities)
        #modify values in node
        NewGene = ModifyGeneNodeValues(NewGene)
        #change a connection (gain new features)
        if random.randint(1,100) <= 20: NewGene=ChangeRandomConnect(NewGene)
        
        GeneratePheno(generation+1,x,NewGene)
        with open(str(generation+1)+"/genes.txt", "a") as myfile:
             myfile.write(str(NewGene))
             myfile.write("\n")
    
def EvolutionLoop(generation,evolutions):
    for x in range(evolutions):
        PerformPheno(generation)
        GetFitness2(generation)
        Breed(generation)
        generation = generation+1


def InteractiveBreed(generation,amount):
    #get genes from generation
    AllGenes= []
    with open (str(generation)+"/genes.txt", "r") as myfile:
        data = myfile.read()
        lines = data.split("\n")
        lines.pop()
        for each_line in lines:
            AllGenes.append(eval(each_line))
            
    #get top entities
    AllScores = []
    with open (str(generation)+"/interactive-fitness.txt", "r") as myfile:
        data = myfile.read()
        AllScores = eval(data)

    TopEntities = []
    for x in range(len(AllScores)):
        TopEntities.append(AllGenes[AllScores[x]])

    print(TopEntities)

    for x in range(amount):
        #sexytime generate new genes from parents
        NewGene=SexyTimeFrom(TopEntities)
        #modify values in node
        NewGene = ModifyGeneNodeValues(NewGene)
        #change a connection 20% (gain new features)
        if random.randint(1,100) <= 20: NewGene=ChangeRandomConnect(NewGene)
        
        GeneratePheno(generation+1,x,NewGene)
        with open(str(generation+1)+"/genes.txt", "a") as myfile:
             myfile.write(str(NewGene))
             myfile.write("\n")

def InteractiveGetFitness(generation):
    txt = input("Fittest entries [x,x,x,x,x]: ")
    with open(str(generation)+"/interactive-fitness.txt", "a") as myfile:
        myfile.write(str(txt))


def InteractiveEvolutionLoop(generation,evolutions,amount):
    for x in range(evolutions):
        PerformPheno(generation)
        InteractiveGetFitness(generation)
        InteractiveBreed(generation,amount)
        generation = generation+1

 
    
def ShowFunctions():
    print('1. GenerateRandomPopulation(gennumber,amount,complexity) \n generates folder with subfolder of entities\n')
    print('AUTO MODE (Object detection) \n')
    print('--> EvolutionLoop(generation,evolutions) \n auto-evolves using object detection model')
    print('INTERACTIVE MODE (prompts user for fitness levels)\n')
    print('--> InteractiveEvolutionLoop(generation,evolutions,amount)')
    
    

ShowFunctions()

#audio=open all files from generation
#spectrogram from all audio files

#GenerateRandomPopulation(2,5,10)
#PerformPheno(2,5)





#quote population



#evolve population


